# Copyright (c) Acconeer AB, 2022
# All rights reserved

import abc
import logging
import multiprocessing as mp
import platform
import queue
import signal
import sys
import traceback
from collections import namedtuple
from time import sleep, time

import numpy as np

from acconeer.exptool import libft4222
from acconeer.exptool.a111._clients import links
from acconeer.exptool.a111._clients.base import (
    BaseClient,
    ClientError,
    SessionSetupError,
    decode_version_str,
)
from acconeer.exptool.a111._clients.reg import protocol, regmap
from acconeer.exptool.a111._modes import Mode


log = logging.getLogger(__name__)

ModeInfo = namedtuple("ModeInfo", ["fixed_buffer_size", "byte_per_element"])

MODE_INFOS = {
    Mode.POWER_BINS: ModeInfo(True, 2),
    Mode.ENVELOPE: ModeInfo(True, 2),
    Mode.IQ: ModeInfo(True, 4),
    Mode.SPARSE: ModeInfo(True, 2),
}


class RegBaseClient(BaseClient):
    _STATUS_TIMEOUT = 3.0

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        self._streaming_control_val = "no_streaming"  # Override in subclass (UART)

        self._mode = None
        self._config = None
        self._data_length = None

    def _setup_session(self, config):
        if len(config.sensor) > 1:
            raise ValueError("the register protocol does not support multiple sensors")
        if config.sensor[0] != 1:
            raise ValueError("the register protocol currently only supports using sensor 1")

        mode = config.mode
        self._mode = mode
        self._config = config

        self._write_reg("main_control", "stop")
        self._write_reg("mode_selection", mode)
        self._write_reg("update_rate", 0)
        self._write_reg("sensor_power_mode", "active")

        for key, reg in regmap.get_config_key_to_reg_map(mode).items():
            val = getattr(config, key)

            if val is None:
                continue

            self._write_reg(reg, val)

        self._write_reg("streaming_control", self._streaming_control_val)

        self._write_reg("main_control", "create")
        self._wait_status(regmap.STATUS_FLAGS.CREATED)

        info = {}
        info_regs = regmap.get_session_info_regs(mode)
        for reg in info_regs:
            k = reg.stripped_name
            k = regmap.STRIPPED_NAME_TO_INFO_REMAP.get(k, k)

            if k is None:
                continue

            info[k] = self._read_reg(reg)

        if MODE_INFOS[mode].fixed_buffer_size:
            self._data_length = info.get("data_length")

            log.info("data length: {}".format(self._data_length))
            log.info("buffer size: {} B".format(self._buffer_size))
        else:
            self._data_length = None

        if self._data_rate is not None:
            log.info("requested data rate: {:.2f} Mbit/s".format(self._data_rate * 1e-6))

        return info

    def _wait_status(self, val, mask=None):
        val = regmap.STATUS_FLAGS(val)

        if mask is None:
            mask = val
        else:
            mask = regmap.STATUS_FLAGS(mask)

        start_time = time()

        while time() - start_time < self._STATUS_TIMEOUT:
            status = self._read_reg(regmap.STATUS_REG)

            status_error = status & regmap.STATUS_MASKS.ERROR_MASK

            if status_error:
                if status_error & regmap.STATUS_FLAGS.ERROR_CREATION:
                    raise SessionSetupError
                else:
                    raise ClientError("server error: " + str(status_error).split(".")[1])

            if (status & mask) == val:
                return

        raise ClientError("timeout while waiting for status")

    def _get_supported_modes(self):
        supported_modes = set()
        self._write_reg("main_control", "clear_status")

        for mode in Mode:
            self._write_reg("mode_selection", mode)
            status = self._read_reg("status")

            if status & regmap.STATUS_FLAGS.ERROR_SET_MODE:
                self._write_reg("main_control", "clear_status")
            else:
                supported_modes.add(mode)

        return supported_modes

    @property
    def _buffer_size(self):  # B
        if self._data_length is None or self._mode is None:
            return None

        return self._data_length * MODE_INFOS[self._mode].byte_per_element

    @property
    def _data_rate(self):  # Mbit/s
        if self._buffer_size is None or self._config.update_rate is None:
            return None

        return self._buffer_size * 8 * self._config.update_rate

    @abc.abstractmethod
    def _read_reg(self, reg):
        pass

    @abc.abstractmethod
    def _write_reg(self, reg, val):
        pass


class UARTClient(RegBaseClient):
    DEFAULT_BASE_BAUDRATE = 115200
    CONNECT_ROUTINE_TIMEOUT = 0.6

    def __init__(self, port, **kwargs):
        self.override_baudrate = kwargs.pop("override_baudrate", None)

        super().__init__(**kwargs)

        self._streaming_control_val = "uart_streaming"

        if platform.system().lower() in ["windows", "darwin"]:
            self._link = links.SerialLink(port)
        else:
            self._link = links.SerialProcessLink(port)

    def _connect(self):
        self._link.timeout = self.CONNECT_ROUTINE_TIMEOUT

        if self.override_baudrate:
            self._link.baudrate = self.override_baudrate
            self._link.connect()

            try:
                self._handshake()
            except links.LinkError as e:
                raise ClientError("could not connect, no response") from e
        else:
            baudrates = [int(3e6), int(1e6)]
            baudrates.append(self.DEFAULT_BASE_BAUDRATE)
            baudrates = sorted(list(set(baudrates)))

            self._link.baudrate = baudrates[0]
            self._link.connect()

            for i, baudrate in enumerate(baudrates):
                if i != 0:
                    self._link.baudrate = baudrate
                    sleep(0.2)

                try:
                    self._handshake()
                except links.LinkError:
                    log.debug("handshake failed at {} baud".format(baudrate))
                else:
                    log.debug("handshake succeeded at {} baud".format(baudrate))
                    break
            else:
                raise ClientError("could not connect, no response")

            product_max_baudrate = self._read_reg("product_max_uart_baudrate")

            if baudrate != product_max_baudrate:
                log.debug("switching to {} baud...".format(product_max_baudrate))
                self._write_reg("uart_baudrate", product_max_baudrate)
                self._link.baudrate = product_max_baudrate
                sleep(0.2)
                self._handshake()
                log.debug("handshake succeeded at {} baud".format(product_max_baudrate))

        self._link.timeout = self._link.DEFAULT_TIMEOUT

        version_buffer = self._read_buf_raw()
        version_info = decode_version_buffer(version_buffer)

        info = {}
        info.update(version_info)

        return info

    def _setup_session(self, config):
        ret = super()._setup_session(config)

        if self._config.update_rate:
            self._link.timeout = 1 / self._config.update_rate + self._link.DEFAULT_TIMEOUT
        else:
            self._link.timeout = self._link.DEFAULT_TIMEOUT

        if self._data_rate is not None:
            if self._data_rate > 2 / 3 * self._link.baudrate:
                log.warning("data rate might be too high")

        return ret

    def _start_session(self):
        self._write_reg("main_control", "activate")
        # self._wait_status(regmap.STATUS_FLAGS.ACTIVATED)
        # TODO: how can we handle streaming and reading/writing registers at the same time?

    def _get_next(self):
        packet = self._recv_packet(allow_recovery_skip=True)

        if not isinstance(packet, protocol.StreamData):
            raise ClientError("got unexpected type of frame")

        info = {}
        for addr, enc_val in packet.result_info:
            try:
                reg = regmap.get_reg(addr, self._mode)
                val = reg.decode(enc_val)
            except (protocol.ProtocolError, ValueError):
                log.info("got unknown reg val in result info")
                log.info("addr: {}, value: {}".format(addr, fmt_enc_val(enc_val)))
            else:
                k = reg.stripped_name
                k = regmap.STRIPPED_NAME_TO_INFO_REMAP.get(k, k)

                if k is None:
                    continue

                info[k] = val

        sweeps_per_frame = getattr(self._config, "sweeps_per_frame", None)
        data = protocol.decode_output_buffer(packet.buffer, self._mode, sweeps_per_frame)

        if self.squeeze:
            return info, data
        else:
            return [info], np.expand_dims(data, 0)

    def _stop_session(self):
        self._write_reg("main_control", "stop", expect_response=False)

        t0 = time()
        while time() - t0 < self._link.timeout:
            res = self._recv_packet()

            if isinstance(res, protocol.RegWriteResponse):
                break
            if not isinstance(res, protocol.StreamData):
                raise ClientError("got unexpected packet while stopping session")
        else:
            raise ClientError("timeout while stopping session")

        self._link.timeout = self._link.DEFAULT_TIMEOUT

        mask = regmap.STATUS_FLAGS.CREATED | regmap.STATUS_FLAGS.ACTIVATED
        self._wait_status(0, mask=mask)

    def _disconnect(self):
        self._write_reg("uart_baudrate", self.DEFAULT_BASE_BAUDRATE)
        self._link.disconnect()

    def _read_buf_raw(self, addr=protocol.MAIN_BUFFER_ADDR):
        req = protocol.BufferReadRequest(addr)
        self._send_packet(req)

        log.debug("sent buf r req: addr: 0x{:02x}".format(addr))

        res = self._recv_packet()
        if not isinstance(res, protocol.BufferReadResponse):
            raise ClientError("got unexpected type of frame")

        log.debug("recv buf r res: addr: 0x{:02x} len: {}".format(addr, len(res.buffer)))

        return res.buffer

    def _read_reg(self, reg, mode=None):
        mode = self._mode if mode is None else mode
        reg = regmap.get_reg(reg, mode)
        enc_val = self._read_reg_raw(reg.addr)
        return reg.decode(enc_val)

    def _read_reg_raw(self, addr):
        addr = regmap.get_reg_addr(addr)
        req = protocol.RegReadRequest(addr)
        self._send_packet(req)

        log.debug("sent reg r req: addr: 0x{:02x}".format(addr))

        res = self._recv_packet()
        if not isinstance(res, protocol.RegReadResponse):
            raise ClientError("got unexpected type of frame")

        enc_val = res.reg_val.val

        log.debug("recv reg r res: addr: 0x{:02x} val: {}".format(addr, fmt_enc_val(enc_val)))

        return enc_val

    def _write_reg(self, reg, val, expect_response=True):
        reg = regmap.get_reg(reg, self._mode)
        enc_val = reg.encode(val)
        self._write_reg_raw(reg.addr, enc_val, expect_response)

    def _write_reg_raw(self, addr, enc_val, expect_response=True):
        addr = regmap.get_reg_addr(addr)
        rrv = protocol.RegVal(addr, enc_val)
        req = protocol.RegWriteRequest(rrv)
        self._send_packet(req)

        log.debug("sent reg w req: addr: 0x{:02x} val: {}".format(addr, fmt_enc_val(enc_val)))

        if expect_response:
            res = self._recv_packet()
            if not isinstance(res, protocol.RegWriteResponse):
                raise ClientError("got unexpected packet (expected reg write response)")
            if res.reg_val != rrv:
                raise ClientError("reg write failed")

            log.debug("recv reg w res: ok")

    def _send_packet(self, packet):
        frame = protocol.insert_packet_into_frame(packet)
        self._link.send(frame)

    def _recv_packet(self, allow_recovery_skip=False):
        buf_1 = self._link.recv(1 + protocol.LEN_FIELD_SIZE)

        start_marker = buf_1[0]
        packet_len = int.from_bytes(buf_1[1:], protocol.BO)

        if start_marker != protocol.START_MARKER:
            raise ClientError("got invalid frame (incorrect start marker)")

        buf_2 = self._link.recv(packet_len + 2)
        packet = buf_2[:-1]
        end_marker = buf_2[-1]

        if end_marker != protocol.END_MARKER:
            if not allow_recovery_skip:
                raise ClientError("got invalid frame (incorrect end marker)")

            log.debug("got invalid frame (incorrect end marker), attempting recovery")

            buf_2.extend(self._link.recv(1 + protocol.LEN_FIELD_SIZE))

            si = 0
            ei = len(buf_2)
            expected_sub_len = protocol.LEN_FIELD_SIZE + packet_len + 3
            while True:
                si = buf_2.find(buf_1, si, ei)
                if si < 0:
                    raise ClientError("got invalid frame and could not recover")

                sub_len_diff = expected_sub_len - (len(buf_2) - si)

                if sub_len_diff > 0:
                    buf_2.extend(self._link.recv(sub_len_diff))

                if buf_2[-1] != protocol.END_MARKER:
                    log.debug("recovery attempt failed")
                    continue

                packet = buf_2[si + 1 + protocol.LEN_FIELD_SIZE : -1]
                break

            log.warning("successfully recovered from corrupt frame")

        return protocol.unpack_packet(packet)

    def _handshake(self):
        self._write_reg("main_control", "stop", expect_response=False)

        exp_addr = regmap.get_reg_addr("main_control")
        exp_enc_val = regmap.get_reg("main_control").encode("stop")
        exp_reg_val = protocol.RegVal(exp_addr, exp_enc_val)
        exp_packet = protocol.RegWriteResponse(exp_reg_val)
        exp_frame = protocol.insert_packet_into_frame(exp_packet)
        self._link.recv_until(exp_frame)

    @property
    def description(self):
        return f"UART ({self._link._port})"


class PollingUARTClient(UARTClient):
    def __init__(self, port, **kwargs):
        self._measure_on_call = kwargs.pop("measure_on_call", True)

        super().__init__(port, **kwargs)
        self._streaming_control_val = "no_streaming"
        self._link = links.SerialLink(port)  # don't use link in separate process

    def _start_session(self):
        self._write_reg("main_control", "activate")
        self._wait_status(regmap.STATUS_FLAGS.ACTIVATED)

    def _get_next(self):
        if self._measure_on_call:
            self._write_reg("main_control", "clear_status")

        poll_t = time()

        while True:
            status = self._read_reg("status")

            if status & regmap.STATUS_MASKS.ERROR_MASK:
                raise ClientError("server error: " + str(status).split(".")[1])
            elif status & regmap.STATUS_FLAGS.DATA_READY:
                break
            else:
                if (time() - poll_t) > self._link.timeout:
                    raise ClientError("gave up polling")

                continue

        buffer = self._read_buf_raw()

        info = {}
        info_regs = regmap.get_data_info_regs(self._config.mode)
        for reg in info_regs:
            k = reg.stripped_name
            k = regmap.STRIPPED_NAME_TO_INFO_REMAP.get(k, k)

            if k is None:
                continue

            info[k] = self._read_reg(reg)

        if not self._measure_on_call:
            self._write_reg("main_control", "clear_status")

        sweeps_per_frame = getattr(self._config, "sweeps_per_frame", None)
        data = protocol.decode_output_buffer(buffer, self._mode, sweeps_per_frame)

        if self.squeeze:
            return info, data
        else:
            return [info], np.expand_dims(data, 0)

    def _stop_session(self):
        self._write_reg("main_control", "stop")

        mask = regmap.STATUS_FLAGS.CREATED | regmap.STATUS_FLAGS.ACTIVATED
        self._wait_status(0, mask=mask)


class SPIClient(RegBaseClient):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        self._proc = None

    def _connect(self):
        self._cmd_queue = mp.Queue()
        self._data_queue = mp.Queue()
        args = (
            self._cmd_queue,
            self._data_queue,
        )
        self._proc = SPICommProcess(*args)
        self._proc.start()

        self.__cmd_proc("connect")

        log.debug("connected")

        version_buffer = self._read_main_buffer()
        version_info = decode_version_buffer(version_buffer)

        info = {}
        info.update(version_info)

        return info

    def _setup_session(self, config):
        ret = super()._setup_session(config)
        update_rate = self._config.update_rate
        self.__cmd_proc("update_state", self._mode, update_rate, self._buffer_size)
        return ret

    def _start_session(self):
        self._write_reg("main_control", "activate")
        self._wait_status(regmap.STATUS_FLAGS.ACTIVATED)

        if self._buffer_size is not None:
            self._wait_status(regmap.STATUS_FLAGS.DATA_READY)
            reported_buffer_size = self._read_reg("output_buffer_length")
            assert self._buffer_size == reported_buffer_size

        self.__cmd_proc("start_session")

    def _get_next(self):
        ret_cmd, ret_args = self._data_queue.get()
        if ret_cmd == "error":
            raise ClientError("exception raised in SPI communcation process")
        elif ret_cmd != "get_next":
            raise ClientError
        info, buffer = ret_args

        sweeps_per_frame = getattr(self._config, "sweeps_per_frame", None)
        data = protocol.decode_output_buffer(buffer, self._mode, sweeps_per_frame)

        if self.squeeze:
            return info, data
        else:
            return [info], np.expand_dims(data, 0)

    def _stop_session(self):
        self.__cmd_proc("stop_session")

        mask = regmap.STATUS_FLAGS.CREATED | regmap.STATUS_FLAGS.ACTIVATED
        self._wait_status(0, mask=mask)

    def _disconnect(self):
        self.__cmd_proc("disconnect")
        self._proc.join(1)

    def _read_reg(self, reg):
        reg = regmap.get_reg(reg, self._mode)
        enc_val = self._read_reg_raw(reg.addr)
        return reg.decode(enc_val)

    def _read_reg_raw(self, addr):
        addr = regmap.get_reg_addr(addr)
        enc_val = self.__cmd_proc("read_reg_raw", addr)
        return enc_val

    def _write_reg(self, reg, val):
        reg = regmap.get_reg(reg, self._mode)
        enc_val = reg.encode(val)
        self._write_reg_raw(reg.addr, enc_val)

    def _write_reg_raw(self, addr, enc_val):
        addr = regmap.get_reg_addr(addr)
        self.__cmd_proc("write_reg_raw", addr, enc_val)

    def read_buf_raw(self, addr, size):
        addr = regmap.get_reg_addr(addr)
        buffer = self.__cmd_proc("read_buf_raw", addr, size)
        return buffer

    def _read_main_buffer(self):
        buffer_size = self._read_reg("output_buffer_length")
        if buffer_size > 0:
            buffer = self.read_buf_raw(protocol.MAIN_BUFFER_ADDR, buffer_size)
        else:
            buffer = bytearray()
        return buffer

    def __cmd_proc(self, cmd, *args):
        log.debug("sending cmd to proc: {}".format(cmd))
        self._cmd_queue.put((cmd, args))
        if cmd == "stop_session":
            while True:
                ret_cmd, _ = self._data_queue.get()
                if ret_cmd == cmd:
                    break
                elif ret_cmd == "error":
                    raise ClientError("exception raised in SPI communcation process")
                elif ret_cmd != "get_next":
                    raise ClientError
            ret_args = None
        else:
            ret_cmd, ret_args = self._data_queue.get()
            if ret_cmd == "error":
                raise ClientError("exception raised in SPI communcation process")
            elif ret_cmd != cmd:
                raise ClientError
        return ret_args

    @property
    def description(self):
        return "SPI"


class SPICommProcess(mp.Process):
    def __init__(self, cmd_q, data_q):
        super().__init__(daemon=True)
        self.cmd_q = cmd_q
        self.data_q = data_q
        self.mode = None

    def run(self):
        self.log = logging.getLogger(__name__)
        self.log.debug("SPI communication process started")
        signal.signal(signal.SIGINT, signal.SIG_IGN)

        try:
            self._run()
        except Exception:
            print("Exception raised in SPI communication process:\n", file=sys.stderr)
            traceback.print_exc()
            print("\n\n", file=sys.stderr)
            self.data_q.put(("error", ()))

        while True:
            try:
                self.cmd_q.get(timeout=0.01)
            except queue.Empty:
                break

    def _run(self):
        while True:
            (cmd, cmd_args) = self.cmd_q.get()

            if hasattr(self, cmd):
                ret = getattr(self, cmd)(*cmd_args)
                self.data_q.put((cmd, ret))

                if cmd == "start_session":
                    self.poll()
                elif cmd == "disconnect":
                    break
            else:
                raise ClientError("unknown cmd {}".format(cmd))

    def poll(self):
        while self.cmd_q.empty():
            ret = self.get_next()
            self.data_q.put(("get_next", ret))

    def get_next(self):
        poll_t = time()

        while True:
            status = self.read_reg("status", do_log=False)

            if status & regmap.STATUS_MASKS.ERROR_MASK:
                raise ClientError("server error: " + str(status).split(".")[1])
            elif status & regmap.STATUS_FLAGS.DATA_READY:
                break
            else:
                if (time() - poll_t) > self.poll_timeout:
                    raise ClientError("gave up polling")

                continue

        buffer_size = self.fixed_buf_size or self.read_reg("output_buffer_length")
        if buffer_size > 0:
            buffer = self.read_buf_raw(protocol.MAIN_BUFFER_ADDR, buffer_size)
        else:
            buffer = bytearray()

        info = {}
        info_regs = regmap.get_data_info_regs(self.mode)
        for reg in info_regs:
            k = reg.stripped_name
            k = regmap.STRIPPED_NAME_TO_INFO_REMAP.get(k, k)

            if k is None:
                continue

            info[k] = self.read_reg(reg, do_log=False)

        self.write_reg("main_control", "clear_status", do_log=False)

        return info, buffer

    def connect(self):
        self.dev = libft4222.Device()
        self.dev.open_ex()
        self.dev.set_clock(libft4222.ClockRate.SYS_CLK_48)
        self.dev.spi_master_init(clock=libft4222.SPIClock.CLK_DIV_2)
        self.dev.spi_set_driving_strength()
        self.dev.set_timeouts(1000, 1000)
        self.dev.set_suspend_out(False)
        self.dev.set_wake_up_interrupt(False)

    def update_state(self, mode, update_rate, buffer_size):
        self.mode = mode

        if update_rate is None:
            self.poll_timeout = 1.0
        else:
            self.poll_timeout = 2 / update_rate + 0.5

        self.fixed_buf_size = buffer_size

    def start_session(self):
        pass

    def stop_session(self):
        self.write_reg("main_control", "stop")

    def disconnect(self):
        self.dev.close()

    def read_reg(self, reg, do_log=True):
        reg = regmap.get_reg(reg, self.mode)
        enc_val = self.read_reg_raw(reg.addr, do_log=do_log)
        return reg.decode(enc_val)

    def read_reg_raw(self, addr, do_log=True):
        addr = regmap.get_reg_addr(addr)
        b = bytearray([protocol.REG_READ_REQUEST, addr, 0, 0])
        self.dev.spi_master_single_write(b)
        enc_val = self.dev.spi_master_single_read(4)
        if do_log:
            log.debug("reg r res: addr: 0x{:02x} val: {}".format(addr, fmt_enc_val(enc_val)))
        return enc_val

    def write_reg(self, reg, val, do_log=True):
        reg = regmap.get_reg(reg, self.mode)
        enc_val = reg.encode(val)
        self.write_reg_raw(reg.addr, enc_val, do_log=do_log)

    def write_reg_raw(self, addr, enc_val, do_log=True):
        addr = regmap.get_reg_addr(addr)
        b = bytearray([protocol.REG_WRITE_REQUEST, addr, 0, 0])
        self.dev.spi_master_single_write(b)
        if do_log:
            log.debug("reg w req: addr: 0x{:02x} val: {}".format(addr, fmt_enc_val(enc_val)))
        self.dev.spi_master_single_write(enc_val)

    def read_buf_raw(self, addr, size):
        b = bytearray([protocol.BUF_READ_REQUEST, addr, 0, 0])
        self.dev.spi_master_single_write(b)
        return self.dev.spi_master_single_read(size)


def fmt_enc_val(enc_val):
    return " ".join(["{:02x}".format(x) for x in enc_val])


def decode_version_buffer(version: bytearray) -> dict:
    try:
        version_str = version.decode("ascii").strip()
        assert len(version_str) > 1
    except (UnicodeDecodeError, AssertionError):
        return {}

    return decode_version_str(version_str)



# Copyright (c) Acconeer AB, 2022
# All rights reserved

from __future__ import annotations

import warnings
from typing import Any, Optional, Tuple

import attrs
import h5py
import numpy as np
from attr import Attribute

from acconeer.exptool import a121
from acconeer.exptool.a121._core.utils import is_divisor_of, is_multiple_of
from acconeer.exptool.a121.algo import AlgoConfigBase

from ._processors import Processor, ProcessorConfig, ProcessorExtraResult


SPARSE_IQ_PPC = 24


@attrs.mutable(kw_only=True)
class DetectorConfig(AlgoConfigBase):
    start_m: float = attrs.field(default=1.0)
    end_m: float = attrs.field(default=2.0)
    step_length: Optional[int] = attrs.field(default=None)
    detection_threshold: float = attrs.field(default=1.5)
    frame_rate: float = attrs.field(default=10.0)
    sweeps_per_frame: int = attrs.field(default=16)
    hwaas: int = attrs.field(default=32)

    @step_length.validator
    def _validate_step_length(self, attrs: Attribute, step_length: int) -> None:
        if step_length is not None:
            if not (
                is_divisor_of(SPARSE_IQ_PPC, step_length)
                or is_multiple_of(SPARSE_IQ_PPC, step_length)
            ):
                raise ValueError(f"step_length must be a divisor or multiple of {SPARSE_IQ_PPC}")


@attrs.frozen(kw_only=True)
class DetectorResult:
    presence_score: float = attrs.field()
    presence_distance: float = attrs.field()
    presence_detected: bool = attrs.field()
    processor_extra_result: ProcessorExtraResult = attrs.field()
    service_result: a121.Result = attrs.field()

    def __str__(self) -> str:
        s = "Presence! " if self.presence_detected else "No presence. "
        s += f"Score {self.presence_score:.3f} at {self.presence_distance:.3f} m "
        return s


class Detector:
    MIN_DIST_M = {
        a121.Profile.PROFILE_1: None,
        a121.Profile.PROFILE_2: 0.28,
        a121.Profile.PROFILE_3: 0.56,
        a121.Profile.PROFILE_4: 0.76,
        a121.Profile.PROFILE_5: 1.28,
    }

    def __init__(
        self,
        *,
        client: a121.Client,
        sensor_id: int,
        detector_config: DetectorConfig,
    ) -> None:
        self.client = client
        self.sensor_id = sensor_id
        self.detector_config = detector_config

        self.started = False

    def start(self, recorder: Optional[a121.Recorder] = None) -> None:
        if self.started:
            raise RuntimeError("Already started")

        sensor_config = self._get_sensor_config(self.detector_config)
        session_config = a121.SessionConfig(
            {self.sensor_id: sensor_config},
            extended=False,
        )

        metadata = self.client.setup_session(session_config)
        assert isinstance(metadata, a121.Metadata)

        processor_config = self._get_processor_config(self.detector_config)

        self.processor = Processor(
            sensor_config=sensor_config,
            metadata=metadata,
            processor_config=processor_config,
        )

        if recorder is not None:
            if isinstance(recorder, a121.H5Recorder):
                algo_group = recorder.require_algo_group("presence_detector")
                _record_algo_data(
                    algo_group,
                    self.sensor_id,
                    self.detector_config,
                )
            else:
                # Should never happen as we currently only have the H5Recorder
                warnings.warn("Will not save algo data")

        self.client.start_session(recorder)

        self.started = True

    @classmethod
    def _get_sensor_config(cls, detector_config: DetectorConfig) -> a121.SensorConfig:
        start_point = int(np.floor(detector_config.start_m / Processor.APPROX_BASE_STEP_LENGTH_M))
        viable_profiles = [
            k for k, v in cls.MIN_DIST_M.items() if v is None or v <= detector_config.start_m
        ]
        profile = viable_profiles[-1]

        if detector_config.step_length is not None:
            step_length = detector_config.step_length
        else:
            fwhm_p = Processor.ENVELOPE_FWHM_M[profile] / Processor.APPROX_BASE_STEP_LENGTH_M
            step_length = int((fwhm_p // SPARSE_IQ_PPC) * SPARSE_IQ_PPC)

        num_point = int(
            np.ceil(
                (detector_config.end_m - detector_config.start_m)
                / (step_length * Processor.APPROX_BASE_STEP_LENGTH_M)
            )
        )
        return a121.SensorConfig(
            profile=profile,
            start_point=start_point,
            num_points=num_point,
            step_length=step_length,
            hwaas=detector_config.hwaas,
            sweeps_per_frame=detector_config.sweeps_per_frame,
            frame_rate=detector_config.frame_rate,
        )

    @classmethod
    def _get_processor_config(cls, detector_config: DetectorConfig) -> ProcessorConfig:
        return ProcessorConfig(
            detection_threshold=detector_config.detection_threshold,
        )

    def get_next(self) -> DetectorResult:
        if not self.started:
            raise RuntimeError("Not started")

        result = self.client.get_next()
        assert isinstance(result, a121.Result)

        assert self.processor is not None
        processor_result = self.processor.process(result)

        return DetectorResult(
            presence_score=processor_result.presence_score,
            presence_distance=processor_result.presence_distance,
            presence_detected=processor_result.presence_detected,
            processor_extra_result=processor_result.extra_result,
            service_result=result,
        )

    def update_config(self, config: DetectorConfig) -> None:
        raise NotImplementedError

    def stop(self) -> Any:
        if not self.started:
            raise RuntimeError("Already stopped")

        recorder_result = self.client.stop_session()

        self.started = False

        return recorder_result


def _record_algo_data(
    algo_group: h5py.Group,
    sensor_id: int,
    detector_config: DetectorConfig,
) -> None:
    algo_group.create_dataset(
        "sensor_id",
        data=sensor_id,
        track_times=False,
    )
    algo_group.create_dataset(
        "detector_config",
        data=detector_config.to_json(),
        dtype=a121._H5PY_STR_DTYPE,
        track_times=False,
    )


def _load_algo_data(algo_group: h5py.Group) -> Tuple[int, DetectorConfig]:
    sensor_id = algo_group["sensor_id"][()]
    config = DetectorConfig.from_json(algo_group["detector_config"][()])
    return sensor_id, config



# Copyright (c) Acconeer AB, 2022
# All rights reserved

from . import a111, utils
from ._structs import configbase
from .pg_process import PGProccessDiedException, PGProcess


try:
    from ._version import __version__
except ImportError:
    __version__ = "0.0.0"



# Copyright (c) Acconeer AB, 2022
# All rights reserved

from __future__ import annotations

from .app_model import Plugin


def load_default_plugins() -> list[Plugin]:
    from acconeer.exptool.a121.algo.distance._detector_plugin import DISTANCE_DETECTOR_PLUGIN
    from acconeer.exptool.a121.algo.presence._detector_plugin import PRESENCE_DETECTOR_PLUGIN
    from acconeer.exptool.a121.algo.sparse_iq._plugin import SPARSE_IQ_PLUGIN
    from acconeer.exptool.a121.algo.virtual_button._plugin import VIRTUAL_BUTTON_PLUGIN

    return [
        SPARSE_IQ_PLUGIN,
        DISTANCE_DETECTOR_PLUGIN,
        PRESENCE_DETECTOR_PLUGIN,
        VIRTUAL_BUTTON_PLUGIN,
    ]



# Copyright (c) Acconeer AB, 2022
# All rights reserved

from .backend_plugin import (
    ExtendedProcessorBackendPluginBase,
    GenericProcessorBackendPluginBase,
    ProcessorBackendPluginBase,
    ProcessorBackendPluginSharedState,
)
from .plot_plugin import (
    ExtendedProcessorPlotPluginBase,
    GenericProcessorPlotPluginBase,
    ProcessorPlotPluginBase,
)
from .plugin import ProcessorPlugin
from .view_plugin import ProcessorViewPluginBase



# Copyright (c) Acconeer AB, 2022
# All rights reserved

from __future__ import annotations

from enum import Enum, auto
from typing import Any, Optional

from attrs import define, field


class Criticality(Enum):
    ERROR = auto()
    WARNING = auto()


@define
class ValidationResult(Exception):
    source: Any = field()
    aspect: Optional[str] = field()
    message: str = field()
    criticality: Criticality = field()

    def __attrs_post_init__(self) -> None:
        super(Exception, self).__init__(self.source, self.aspect, self.message, self.criticality)

    def __str__(self):
        return self.message


@define
class ValidationError(ValidationResult):
    criticality: Criticality = field(default=Criticality.ERROR, init=False)


@define
class ValidationWarning(ValidationResult, Warning):
    criticality: Criticality = field(default=Criticality.WARNING, init=False)

    def __attrs_post_init__(self) -> None:
        super(Warning, self).__init__(self.source, self.aspect, self.message, self.criticality)



# Copyright (c) Acconeer AB, 2022
# All rights reserved

from __future__ import annotations

import platform

import qtawesome as qta

from PySide6.QtCore import QSize
from PySide6.QtWidgets import (
    QComboBox,
    QHBoxLayout,
    QLabel,
    QLineEdit,
    QPushButton,
    QStackedWidget,
    QWidget,
)

from acconeer.exptool.app.new._enums import ConnectionInterface, ConnectionState
from acconeer.exptool.app.new.app_model import AppModel
from acconeer.exptool.app.new.qt_subclasses import AppModelAwareWidget

from .misc import BUTTON_ICON_COLOR, SerialPortComboBox, USBDeviceComboBox


class _ConnectAndDisconnectButton(QPushButton):
    def __init__(self, app_model: AppModel, parent: QWidget) -> None:
        super().__init__(parent)

        self.app_model = app_model

        self.setFixedWidth(120)

        app_model.sig_notify.connect(self._on_app_model_update)

        self.clicked.connect(self._on_click)

    def _on_click(self) -> None:
        if self.app_model.connection_state == ConnectionState.DISCONNECTED:
            self.app_model.connect_client()
        elif self.app_model.connection_state == ConnectionState.CONNECTED:
            self.app_model.disconnect_client()
        else:
            raise RuntimeError

    def _on_app_model_update(self, app_model: AppModel) -> None:
        TEXTS = {
            ConnectionState.DISCONNECTED: "Connect",
            ConnectionState.CONNECTING: "Connecting...",
            ConnectionState.CONNECTED: "Disconnect",
            ConnectionState.DISCONNECTING: "Disconnecting...",
        }
        ENABLED_STATES = {ConnectionState.CONNECTED, ConnectionState.DISCONNECTED}
        ICONS = {
            ConnectionState.DISCONNECTED: "fa5s.link",
            ConnectionState.CONNECTING: "fa5s.link",
            ConnectionState.CONNECTED: "fa5s.unlink",
            ConnectionState.DISCONNECTING: "fa5s.unlink",
        }

        self.setText(TEXTS[app_model.connection_state])
        self.setEnabled(
            app_model.connection_state in ENABLED_STATES
            and app_model.plugin_state.is_steady
            and app_model.is_connect_ready()
        )
        self.setIcon(qta.icon(ICONS[app_model.connection_state], color=BUTTON_ICON_COLOR))


class _SocketConnectionWidget(AppModelAwareWidget):
    def __init__(self, app_model: AppModel, parent: QWidget) -> None:
        super().__init__(app_model, parent)
        self.app_model = app_model

        self.setLayout(QHBoxLayout(self))
        self.layout().setContentsMargins(0, 0, 0, 0)

        self.ip_line_edit = QLineEdit(self)
        self.ip_line_edit.setPlaceholderText("<IP address>")
        self.ip_line_edit.editingFinished.connect(self._on_line_edit)
        self.ip_line_edit.setMinimumWidth(125)
        self.layout().addWidget(self.ip_line_edit)

    def _on_line_edit(self) -> None:
        self.app_model.set_socket_connection_ip(self.ip_line_edit.text())


class ClientConnectionWidget(AppModelAwareWidget):
    def __init__(self, app_model: AppModel, parent: QWidget) -> None:
        super().__init__(app_model, parent)

        self.app_model = app_model

        self.setLayout(QHBoxLayout(self))
        self.layout().setContentsMargins(0, 0, 0, 0)

        self.interface_dd = QComboBox(self)
        self.layout().addWidget(self.interface_dd)

        self.interface_dd.addItem("Socket", userData=ConnectionInterface.SOCKET)
        self.interface_dd.addItem("Serial", userData=ConnectionInterface.SERIAL)

        if platform.system().lower() == "windows":
            self.interface_dd.addItem("USB", userData=ConnectionInterface.USB)

        self.interface_dd.currentIndexChanged.connect(self._on_interface_dd_change)

        self.stacked = QStackedWidget(self)
        self.stacked.setStyleSheet("QStackedWidget {background-color: transparent;}")
        self.stacked.addWidget(_SocketConnectionWidget(app_model, self.stacked))
        self.stacked.addWidget(SerialPortComboBox(app_model, self.stacked))
        if platform.system().lower() == "windows":
            self.stacked.addWidget(USBDeviceComboBox(app_model, self.stacked))
        self.layout().addWidget(self.stacked)

        self.layout().addWidget(_ConnectAndDisconnectButton(app_model, self))

        self.warning_label = QLabel(self)
        self.warning_label.setPixmap(qta.icon("fa.warning", color="red").pixmap(QSize(16, 16)))
        self.layout().addWidget(self.warning_label)
        self.warning_label.hide()

        self.layout().addStretch(1)

    def _on_interface_dd_change(self) -> None:
        self.app_model.set_connection_interface(self.interface_dd.currentData())

    def on_app_model_update(self, app_model: AppModel) -> None:
        self.stacked.setEnabled(
            app_model.connection_state == ConnectionState.DISCONNECTED,
        )
        self.interface_dd.setEnabled(
            app_model.connection_state == ConnectionState.DISCONNECTED,
        )

        interface_index = self.interface_dd.findData(app_model.connection_interface)
        if interface_index == -1:
            raise RuntimeError

        self.interface_dd.setCurrentIndex(interface_index)
        self.stacked.setCurrentIndex(interface_index)

        if app_model.connection_warning:
            self.warning_label.setToolTip(app_model.connection_warning)
            self.warning_label.show()
        else:
            self.warning_label.hide()


class GenerationSelection(QComboBox):
    def __init__(self, app_model: AppModel, parent: QWidget) -> None:
        super().__init__(parent)

        self.addItem("A121")
        self.setEnabled(False)



# Copyright (c) Acconeer AB, 2022
# All rights reserved

from acconeer.exptool import a121


# Client is an object that is used to interact with the sensor.
client = a121.Client(
    # ip_address="<ip address of host (like a RPi). E.g. 192.168.XXX.YYY>",
    # or
    # serial_port="<serial port of module. E.g. COM3 or /dev/ttyUSBx for Window/Linux>",
)

# Establishes a connection to the server running on the sensor.
client.connect()

# Once the client is connected, information about the server can be accessed.
print("Server Info:")
print(client.server_info)

# In order to get radar data from the server, we need to start a session.

# To be able to start a session, we must first configure the session
sensor_config = a121.SensorConfig()
sensor_config.num_points = 6
sensor_config.sweeps_per_frame = 4
sensor_config.hwaas = 16
client.setup_session(sensor_config)

# Now we are ready to start it:
client.start_session()

n = 5
for i in range(n):
    # Data is retrieved from the sensor with "get_next".
    result = client.get_next()

    print(f"Result {i + 1}:")
    print(result)

# When we are done, we should close the connection to the server.
client.disconnect()



# Copyright (c) Acconeer AB, 2022
# All rights reserved

from __future__ import annotations

import importlib
import logging
import platform
import traceback
from typing import Optional

import qtawesome as qta
import serial

from PySide6.QtCore import Qt, QThread, Signal
from PySide6.QtGui import QCloseEvent, QMovie
from PySide6.QtWidgets import (
    QComboBox,
    QDialog,
    QFileDialog,
    QFormLayout,
    QLabel,
    QLineEdit,
    QPushButton,
    QSizePolicy,
    QStackedWidget,
    QVBoxLayout,
    QWidget,
)

from acconeer.exptool.app import resources  # type: ignore[attr-defined]
from acconeer.exptool.app.new._enums import ConnectionInterface, ConnectionState
from acconeer.exptool.app.new.app_model import AppModel
from acconeer.exptool.app.new.ui.misc import ExceptionWidget
from acconeer.exptool.flash import find_flash_port, flash_image  # type: ignore[import]

from .misc import BUTTON_ICON_COLOR, SerialPortComboBox, USBDeviceComboBox


log = logging.getLogger(__name__)


class _FlashThread(QThread):
    flash_failed = Signal(Exception, str)
    flash_done = Signal()

    def __init__(self, bin_file: str, flash_port: serial.tools.list_ports.ListPortInfo) -> None:
        super().__init__()
        self.bin_file = bin_file
        self.flash_port = flash_port

    def run(self) -> None:
        try:
            flash_image(self.bin_file, self.flash_port)
            self.flash_done.emit()
        except Exception as e:
            log.error(str(e))
            self.flash_failed.emit(e, traceback.format_exc())


class _FlashDialog(QDialog):
    def __init__(self, parent: QWidget) -> None:
        super().__init__(parent)

        self.setWindowTitle("Flash tool")
        self.setMinimumWidth(250)

        vbox = QVBoxLayout(self)
        vbox.setAlignment(Qt.AlignCenter)

        self.loading = QLabel()
        self.loading.setAlignment(Qt.AlignCenter)

        loader_gif = None
        with importlib.resources.path(resources, "loader.gif") as path:
            loader_gif = path

        self.flash_movie = QMovie(str(loader_gif))
        self.loading.setMovie(self.flash_movie)
        vbox.addWidget(self.loading)

        self.flash_label = QLabel(self)
        self.flash_label.setSizePolicy(QSizePolicy.Preferred, QSizePolicy.MinimumExpanding)
        self.flash_label.setAlignment(Qt.AlignCenter)
        vbox.addWidget(self.flash_label)

        self.setLayout(vbox)

    def flash(self, bin_file, flash_port):
        self.flash_thread = _FlashThread(bin_file, flash_port)
        self.flash_thread.started.connect(self._flash_start)
        self.flash_thread.finished.connect(self.flash_thread.deleteLater)
        self.flash_thread.finished.connect(self._flash_stop)
        self.flash_thread.flash_done.connect(self._flash_done)
        self.flash_thread.flash_failed.connect(self._flash_failed)

        self.flash_thread.start()
        self._flashing = True
        self.exec()

    def _flash_start(self) -> None:
        self.flash_label.setText("Flashing...")
        self.loading.show()
        self.flash_movie.start()

    def _flash_stop(self) -> None:
        self.flash_movie.stop()
        self.loading.hide()
        self._flashing = False

    def _flash_done(self) -> None:
        self.flash_label.setText("Flashing done!")

    def _flash_failed(self, exception: Exception, traceback_str: Optional[str]) -> None:
        self.flash_label.setText("Flashing failed!")
        ExceptionWidget(self, exc=exception, traceback_str=traceback_str).exec()

    def closeEvent(self, event: QCloseEvent) -> None:
        if self._flashing:
            self.flash_thread.terminate()
            self.flash_thread.wait()
        super().closeEvent(event)


class _FlashPopup(QDialog):
    def __init__(self, app_model: AppModel, parent: QWidget) -> None:
        super().__init__(parent)

        self.app_model = app_model

        self.setWindowTitle("Flash tool")
        self.setMinimumWidth(350)

        self.bin_file = None

        layout = QFormLayout(self)

        self.file_label = QLineEdit(self)
        self.file_label.setReadOnly(True)
        self.file_label.setPlaceholderText("<Select a bin file>")

        browse_button = QPushButton("Browse", self)
        browse_button.clicked.connect(self._browse_file)
        layout.addRow(browse_button, self.file_label)

        self.interface_dd = QComboBox(self)
        self.interface_dd.addItem("Serial", userData=ConnectionInterface.SERIAL)
        if platform.system().lower() == "windows":
            self.interface_dd.addItem("USB", userData=ConnectionInterface.USB)

        self.interface_dd.currentIndexChanged.connect(self._on_interface_dd_change)

        self.stacked = QStackedWidget(self)
        self.stacked.setStyleSheet("QStackedWidget {background-color: transparent;}")
        self.stacked.addWidget(SerialPortComboBox(app_model, self.stacked))
        if platform.system().lower() == "windows":
            self.stacked.addWidget(USBDeviceComboBox(app_model, self.stacked))

        layout.addRow(self.interface_dd, self.stacked)

        self.flash_button = QPushButton("Flash", self)
        self.flash_button.clicked.connect(self._flash)
        self.flash_button.setEnabled(False)
        layout.addRow(self.flash_button)

        self.setLayout(layout)

        self.browse_file_dialog = QFileDialog(None)
        self.browse_file_dialog.setNameFilter("Bin files (*.bin)")

        self.flash_dialog = _FlashDialog(self)

        app_model.sig_notify.connect(self._on_app_model_update)

    @property
    def flash_port(self) -> Optional[str]:
        if self.app_model.connection_interface == ConnectionInterface.SERIAL:
            return self.app_model.serial_connection_port
        if self.app_model.connection_interface == ConnectionInterface.USB:
            return self.app_model.usb_connection_device
        return None

    def _browse_file(self) -> None:
        if self.browse_file_dialog.exec():
            filenames = self.browse_file_dialog.selectedFiles()
            self.bin_file = filenames[0]

        self._draw()

    def _flash(self) -> None:
        flash_port = find_flash_port(self.flash_port)

        self.flash_dialog.flash(self.bin_file, flash_port)

    def _on_interface_dd_change(self) -> None:
        self.app_model.set_connection_interface(self.interface_dd.currentData())

    def _on_app_model_update(self, app_model: AppModel) -> None:
        interface_index = 0
        if app_model.connection_interface in [ConnectionInterface.SERIAL, ConnectionInterface.USB]:
            interface_index = self.interface_dd.findData(app_model.connection_interface)
            if interface_index == -1:
                raise RuntimeError

        self.interface_dd.setCurrentIndex(interface_index)
        self.stacked.setCurrentIndex(interface_index)

        self._draw()

    def _draw(self) -> None:
        self.file_label.setText(self.bin_file if self.bin_file else "")
        self.flash_button.setEnabled(self.flash_port is not None and self.bin_file is not None)

    def exec(self) -> None:
        if self.app_model.connection_interface == ConnectionInterface.SOCKET:
            self.app_model.set_connection_interface(ConnectionInterface.SERIAL)

        super().exec()


class FlashButton(QPushButton):
    def __init__(self, app_model: AppModel, parent: QWidget) -> None:
        super().__init__(parent)

        self.app_model = app_model

        self.setFixedWidth(100)
        self.setText("Flash")
        self.setIcon(qta.icon("mdi.flash", color=BUTTON_ICON_COLOR))

        app_model.sig_notify.connect(self._on_app_model_update)
        self.pop_up = _FlashPopup(app_model, self)
        self.clicked.connect(self._on_click)

    def _on_click(self) -> None:
        self.pop_up.exec()

    def _on_app_model_update(self, app_model: AppModel) -> None:
        self.setEnabled(app_model.connection_state == ConnectionState.DISCONNECTED)



# Copyright (c) Acconeer AB, 2022
# All rights reserved

import argparse
import configparser
import datetime
import fnmatch
import re
import subprocess
import sys
from enum import Enum, auto
from pathlib import Path


class Status(Enum):
    """ """

    OK = auto()
    NOT_SUPPORTED = auto()
    MISSING = auto()
    YEAR_NOT_UP_TO_DATE = auto()


def update_copyright_year(file_content, copyright_line_nbr, current_year, copyright_pattern):
    """
    Update the year in a copyright statement in a source to the current year.
    If the current year is already present, nothing will be done.
    If no year is stated the result is undefined behavior.

    file_content   - The content of the source file as a list where every entry is a line
    copyright_line - The line where the year is stated (first line has index 0)
    """

    if copyright_line_nbr is None:
        copyright_statement = [copyright_pattern[0] + "\n"] + [copyright_pattern[1] + "\n\n"]
        file_content = copyright_statement + file_content
    else:
        copyright_line = file_content[copyright_line_nbr]

        # If the current year is in the file, the job has been done for us
        if str(current_year) in copyright_line:
            return

        # We use regex to find the year in copyright statement
        year_format = re.compile(r".*([1-2][0-9]{3})")
        result = year_format.match(copyright_line)
        start_pos = result.start(1)
        end_pos = result.end(1)

        # Check whether we had a standalone year, e.g., 2017, or a range, e.g, 2017-2018
        # In either case update the year to the current year
        if copyright_line[start_pos - 1] == "-":
            file_content[copyright_line_nbr] = (
                copyright_line[:start_pos]
                + str(current_year)
                + copyright_line[start_pos + 4 : end_pos]
                + "\n"
            )
        else:
            file_content[copyright_line_nbr] = (
                copyright_line[: start_pos + 4]
                + "-"
                + str(current_year)
                + copyright_line[start_pos + 5 : end_pos + 4]
                + "\n"
            )

    return file_content


def match_copyright_pattern(file_content, patterns):
    """
    Match a list of patterns against the file content.
    Returns the line number where the first matching pattern begins. If no match exists,
    None is returned.

    file_content - The content of the source file as a list where every entry is a line
    patterns - A list of patterns (a pattern is a list of strings)
    """
    if not all(isinstance(p, list) for p in patterns):
        patterns = [patterns]

    if any([len(file_content) >= len(p) for p in patterns]):
        for row, line in enumerate(file_content):
            for pattern in patterns:
                if pattern[0] in line:
                    line_matches = [i in j for i, j in zip(pattern, file_content[row:])]
                    if all(line_matches):
                        return row

    return None


def year_string_to_range(s):
    """
    Convert a year string to a list of years containing all years.
    '2022' -> [2022]
    '2020-2022' -> [2020, 2021, 2022]
    """
    year_range = sum(
        (
            (
                list(range(*[int(j) + k for k, j in enumerate(i.split("-"))]))
                if "-" in i
                else [int(i)]
            )
            for i in s.split(",")
        ),
        [],
    )

    return sorted(set(year_range))


def match_copyright_year(line, year_pattern, year):
    """
    Check if the line matches the regex pattern and year, none is returned if regex match fails
    or the year is not matching.

    line         - The line where the year is stated
    year_pattern - The copyright year line regex pattern
    year         - The year that should be checked for
    """
    years_match = year_pattern.match(line)

    if years_match:
        years = years_match.group("date")
        return year in year_string_to_range(years)

    return None


def copyright_pattern_py(year=""):
    """
    Copyright pattern for python files

    year - Replaces year in the copyright statement.
            Should be left empty when used to match against file content.
    """
    return [f"# Copyright (c) Acconeer AB, {year}", "# All rights reserved"]


def copyright_exists_py(file_content):
    """
    Check if a copyright statement is present in a .py file and is in the correct format.
    Returns the line number where the copyright statement begins. If no statement exists or
    is in the wrong format, None is returned.

    file_content - The content of the source file as a list where every entry is a line
    """

    return match_copyright_pattern(file_content, copyright_pattern_py())


def copyright_check_year_py(file_content, line_nbr, current_year):
    """
    Check that the current year is present in a copyright statement in a .py-file.

    file_content - the content of the file
    line_nbr     - the line number where the copyright statement begins
    """

    year_pattern = re.compile(
        r"# Copyright (\(c\)|&copy;) Acconeer AB, "
        + r"(?P<date>\d+(-\d+)?)(. All rights reserved.)?\n"
    )

    return match_copyright_year(file_content[line_nbr], year_pattern, current_year)


def read_file(file_path):
    """
    Read a source file and return its content

    file_path - the absolute or relative path to the source file
    """

    contents = ""
    with open(file_path, "r", errors=None) as file:
        contents = list(file)

    return contents


def write_file(file_path, content):
    """
    Write to a file.

    file_path - the absolute or relative path to the source file
    content   - the content to be written
    """

    content = "".join(content)
    with open(file_path, "w") as header_file:
        header_file.write(content)


def check_copyright(file_path, ignore_year, update, current_year):
    """
    Check copyright statement of a file

    file_path    - The file to check
    ignore_year  - Don't check the year stated in copyright statement
    update       - Update year of copyright statement or add copyright statement if missing
    current_year - The current year. Used to check year of copyright statement
    """

    link_copyright_methods = {
        ".py": (copyright_exists_py, copyright_check_year_py, copyright_pattern_py),
    }

    copyright_methods = link_copyright_methods.get(Path(file_path).suffix, None)

    if copyright_methods:
        copyright_exist, copyright_check_year, copyright_pattern = copyright_methods

        # Read the file
        try:
            content = read_file(file_path)
        except ValueError:
            return (
                Status.NOT_SUPPORTED,
                "{}: encoding seems to be something else than utf-8".format(file_path),
            )

        # Empty files without copyright is accepted
        if len(content):
            # Check if there is a correctly formatted copyright statement
            copyright_line_nbr = copyright_exist(content)

            if copyright_line_nbr is None and not update:
                return (
                    Status.MISSING,
                    "{}: copyright statement is missing or has incorrect format.".format(
                        file_path
                    ),
                )

            if not current_year:
                current_year = datetime.datetime.now().year

            year_up_to_date = (
                copyright_check_year(content, copyright_line_nbr, current_year)
                if copyright_line_nbr is not None
                else False
            )

            # Update the year and write it to the source file
            if update and not year_up_to_date:
                content = update_copyright_year(
                    content, copyright_line_nbr, current_year, copyright_pattern(current_year)
                )
                write_file(file_path, content)
                print("{}: updated copyright year.".format(file_path))

            # Check that the copyright year is up to date
            elif not ignore_year and not year_up_to_date:
                return Status.YEAR_NOT_UP_TO_DATE, "{}: copyright year is not up-to-date.".format(
                    file_path
                )

        return Status.OK, None
    else:
        return Status.NOT_SUPPORTED, None


def check_copyright_files(file_paths, ignore_year=False, update_year=False, current_year=None):
    """
    Check that Acconeer copyright statement is present and correctly formatted in source files.
    A list containing files with incorrect copyright statement is returned.

    file_paths  - a list containing file paths
    ignore_year - set to True if the copyright year should be ignored when checking the format
    update_year - if True, the year in the copyright statement will be updated to the current year
    """

    incorrect_files = []
    for file_path in file_paths:
        file_status, print_str = check_copyright(file_path, ignore_year, update_year, current_year)
        if print_str:
            print(print_str)
        if file_status not in {Status.OK, Status.NOT_SUPPORTED}:
            incorrect_files.append(file_path)

    return incorrect_files


def _get_modified_files():
    # This will show all modified files except deleted and wholly renamed files
    git_args = [
        "git",
        "show",
        "--name-only",
        "--diff-filter=dr",
        "-M100%",
        "--format=tformat:",
        "HEAD",
    ]
    completed_process = subprocess.run(
        git_args, stdout=subprocess.PIPE, universal_newlines=True, check=False
    )
    completed_process.check_returncode()

    return completed_process.stdout.splitlines()


def _get_all_files():
    git_args = ["git", "ls-files"]
    completed_process = subprocess.run(
        git_args, stdout=subprocess.PIPE, universal_newlines=True, check=False
    )
    completed_process.check_returncode()

    return completed_process.stdout.splitlines()


def _get_commit_time():
    git_args = [
        "git",
        "log",
        "-1",
        "--date=unix",
        "--format=%cd",
        "HEAD",
    ]
    completed_process = subprocess.run(
        git_args, stdout=subprocess.PIPE, universal_newlines=True, check=False
    )
    completed_process.check_returncode()

    return int(completed_process.stdout.splitlines()[0])


def main():
    """Main entry function"""

    arg_parser = argparse.ArgumentParser()
    arg_parser.add_argument(
        "-a",
        "--all-files",
        action="store_true",
        help="check all repository files, not just the changed ones",
    )

    arg_parser.add_argument("--year", type=int, help="year to use for copyright year check")

    arg_parser.add_argument(
        "-u", "--update-year", action="store_true", help="update the year in the copyright"
    )

    args = arg_parser.parse_args()
    config = configparser.ConfigParser()
    config.read("setup.cfg")
    section = "check_copyright"

    ignore_files = config.get(section, "ignore_files", fallback="")
    ignore_files = [s.strip() for s in ignore_files.split(",")]
    ignore_files = [s for s in ignore_files if s]

    current_year = args.year

    if not current_year:
        commit_time = _get_commit_time()

        current_year = (
            datetime.datetime.fromtimestamp(commit_time, tz=datetime.timezone.utc).year
            if commit_time
            else None
        )

    filenames = _get_modified_files() if not args.all_files else _get_all_files()

    for filename in filenames.copy():
        if any(fnmatch.fnmatch(filename, p) for p in ignore_files):
            filenames.remove(filename)

    incorrect_files = check_copyright_files(
        filenames, update_year=args.update_year, current_year=current_year
    )

    # Check if there were incorrect files
    if incorrect_files:
        return 1

    return 0


if __name__ == "__main__":
    sys.exit(main())



# Copyright (c) Acconeer AB, 2022
# All rights reserved

from __future__ import annotations

from typing import Optional

import importlib_metadata
import numpy as np

from PySide6 import QtCore
from PySide6.QtWidgets import (
    QFrame,
    QHBoxLayout,
    QLabel,
    QMainWindow,
    QScrollArea,
    QSplitter,
    QStatusBar,
    QVBoxLayout,
    QWidget,
)

from acconeer.exptool.app.new.app_model import AppModel

from .connection_widget import ClientConnectionWidget, GenerationSelection
from .flash_widget import FlashButton
from .misc import ConnectionHint, ExceptionWidget, VerticalSeparator
from .plugin_widget import PluginControlArea, PluginPlotArea, PluginSelection
from .recording_widget import RecordingWidget


class MainWindow(QMainWindow):
    def __init__(self, app_model: AppModel) -> None:
        super().__init__()

        self.resize(1280, 720)

        self.setCentralWidget(MainWindowCentralWidget(app_model, self))
        self.setStatusBar(StatusBar(app_model, self))
        self.setWindowTitle("Acconeer Exploration Tool (Beta)")

        app_model.sig_error.connect(self.on_app_model_error)

    def on_app_model_error(self, exception: Exception, traceback_str: Optional[str]) -> None:
        ExceptionWidget(self, exc=exception, traceback_str=traceback_str).exec()


class MainWindowCentralWidget(QWidget):
    def __init__(self, app_model: AppModel, parent: QWidget) -> None:
        super().__init__(parent)

        self.setLayout(QVBoxLayout(self))
        self.layout().setContentsMargins(0, 0, 0, 0)
        self.layout().setSpacing(0)

        top_bar = TopBar(app_model, self)
        self.layout().addWidget(top_bar)
        self.layout().setStretchFactor(top_bar, 0)

        working_area = WorkingArea(app_model, self)
        self.layout().addWidget(working_area)
        self.layout().setStretchFactor(working_area, 1)


class TopBar(QFrame):
    def __init__(self, app_model: AppModel, parent: QWidget) -> None:
        super().__init__(parent)

        self.setObjectName("TopBar")
        self.setStyleSheet("QFrame#TopBar {background: #ebebeb;}")

        self.setLayout(QHBoxLayout(self))

        self.layout().addWidget(GenerationSelection(app_model, self))
        self.layout().addWidget(VerticalSeparator(self))
        self.layout().addWidget(ClientConnectionWidget(app_model, self))
        self.layout().addWidget(VerticalSeparator(self))
        self.layout().addWidget(FlashButton(app_model, self))
        self.layout().addWidget(ConnectionHint(app_model, self))
        self.layout().addStretch(1)
        self.layout().addWidget(RecordingWidget(app_model, self))


class WorkingArea(QSplitter):
    def __init__(self, app_model: AppModel, parent: QWidget) -> None:
        super().__init__(parent)

        self.setLayout(QHBoxLayout(self))
        self.layout().setContentsMargins(0, 0, 0, 0)
        self.layout().setSpacing(0)

        left_area = LeftArea(app_model, self)
        self.layout().addWidget(left_area)
        self.layout().setStretchFactor(left_area, 0)

        plot_plugin_area = PluginPlotArea(app_model, self)
        self.layout().addWidget(plot_plugin_area)
        self.layout().setStretchFactor(plot_plugin_area, 1)

        right_area = RightArea(app_model, self)
        self.layout().addWidget(right_area)
        self.layout().setStretchFactor(right_area, 0)


class LeftArea(QScrollArea):
    def __init__(self, app_model: AppModel, parent: QWidget) -> None:
        super().__init__(parent)

        self.setWidgetResizable(True)
        self.setFrameShape(QFrame.NoFrame)
        self.setHorizontalScrollBarPolicy(QtCore.Qt.ScrollBarAlwaysOff)
        self.horizontalScrollBar().setEnabled(False)

        self.setMinimumWidth(250)
        self.setMaximumWidth(350)

        self.setWidget(LeftAreaContent(app_model, self))


class LeftAreaContent(QWidget):
    def __init__(self, app_model: AppModel, parent: QWidget) -> None:
        super().__init__(parent)

        self.setLayout(QVBoxLayout(self))

        self.layout().addWidget(PluginSelection(app_model, self))

        self.layout().addStretch(1)


class RightArea(QScrollArea):
    def __init__(self, app_model: AppModel, parent: QWidget) -> None:
        super().__init__(parent)

        self.setWidgetResizable(True)
        self.setFrameShape(QFrame.NoFrame)
        self.setHorizontalScrollBarPolicy(QtCore.Qt.ScrollBarAlwaysOff)
        self.horizontalScrollBar().setEnabled(False)

        self.setMinimumWidth(350)
        self.setMaximumWidth(400)

        self.setWidget(RightAreaContent(app_model, self))


class RightAreaContent(QWidget):
    def __init__(self, app_model: AppModel, parent: QWidget) -> None:
        super().__init__(parent)

        self.setLayout(QVBoxLayout(self))

        self.layout().addWidget(PluginControlArea(app_model, self))


class UpdateRateLabel(QLabel):
    _FPS: int = 60

    def __init__(self, app_model: AppModel, parent: QWidget) -> None:
        super().__init__(parent)

        self.setToolTip("Reported update rate / jitter")

        self.rate = np.nan
        self.jitter = np.nan

        self.startTimer(int(1000 / self._FPS))

        app_model.sig_update_rate.connect(self._on_app_model_update_rate)

    def timerEvent(self, event: QtCore.QTimerEvent) -> None:
        if np.isnan(self.rate) or np.isnan(self.jitter):
            css = "color: #888;"
            text = "- Hz / - ms"
        else:
            css = ""
            text = f"{self.rate:>6.1f} Hz / {self.jitter * 1e3:5.1f} ms"

        self.setStyleSheet(css)
        self.setText(text)

    def _on_app_model_update_rate(self, rate: float, jitter: float) -> None:
        self.rate = rate
        self.jitter = jitter


class StatusBar(QStatusBar):
    def __init__(self, app_model: AppModel, parent: QWidget) -> None:
        super().__init__(parent)

        app_model.sig_notify.connect(self._on_app_model_update)
        app_model.sig_status_message.connect(self._on_app_model_status_message)

        self.message_timer = QtCore.QTimer(self)
        self.message_timer.setInterval(3000)
        self.message_timer.setSingleShot(True)
        self.message_timer.timeout.connect(self._on_timer)

        self.message_widget = QLabel(self)
        self.addWidget(self.message_widget)

        self.addPermanentWidget(UpdateRateLabel(app_model, self))

        self.rss_version_label = QLabel(self)
        self.addPermanentWidget(self.rss_version_label)

        et_version = importlib_metadata.version("acconeer-exptool")
        et_version_text = f"ET: {et_version}"
        self.addPermanentWidget(QLabel(et_version_text, self))

    def _on_app_model_update(self, app_model: AppModel) -> None:
        if app_model.rss_version is None:
            css = "color: #888;"
            text = "RSS: <not connected>"
        else:
            css = ""
            text = f"RSS: {app_model.rss_version}"

        self.rss_version_label.setStyleSheet(css)
        self.rss_version_label.setText(text)

    def _on_app_model_status_message(self, message: Optional[str]) -> None:
        self.message_timer.stop()

        if message:
            self.message_widget.setText(message)
            self.message_timer.start()
        else:
            self.message_widget.clear()

    def _on_timer(self) -> None:
        self.message_widget.clear()



# Copyright (c) Acconeer AB, 2022
# All rights reserved

from __future__ import annotations

from typing import Optional

import attrs

from ._core import IdleState, Metadata, SensorConfig, SessionConfig


@attrs.frozen
class _PerformanceCalc:
    """Calculates performance metrics from config and metadata

    The figures are preliminary, approximate, and not guaranteed. Based on conservative
    measurements taken in room temperature.
    """

    session_config: SessionConfig = attrs.field()
    metadata: Optional[Metadata] = attrs.field()

    @property
    def sensor_config(self) -> SensorConfig:
        return self.session_config.sensor_config

    @property
    def spf(self) -> int:
        return self.sensor_config.sweeps_per_frame

    @property
    def sweep_rate(self) -> float:
        if self.sensor_config.sweep_rate is not None:
            return self.sensor_config.sweep_rate
        else:
            if self.metadata is None:
                raise ValueError("Metadata is None")
            return self.metadata.max_sweep_rate

    @property
    def sweep_period(self) -> float:
        return 1 / self.sweep_rate

    @property
    def sweep_active_duration(self) -> float:
        if self.metadata is None:
            raise ValueError("Metadata is None")
        return 1 / self.metadata.max_sweep_rate

    @property
    def sweep_idle_duration(self) -> float:
        return self.sweep_period - self.sweep_active_duration

    @property
    def frame_active_duration(self) -> float:
        return self.sweep_period * self.spf

    @property
    def frame_rate(self) -> float:
        if self.session_config.update_rate is not None:
            return self.session_config.update_rate

        if self.sensor_config.frame_rate is not None:
            return self.sensor_config.frame_rate

        return self.sweep_rate / self.spf

    @property
    def frame_period(self) -> float:
        return 1 / self.frame_rate

    @property
    def frame_idle_duration(self) -> float:
        return self.frame_period - self.frame_active_duration

    @property
    def active_current(self) -> float:
        return 85e-3

    @property
    def inter_sweep_idle_current(self) -> float:
        return self._get_idle_current(self.sensor_config.inter_sweep_idle_state)

    @property
    def inter_frame_idle_current(self) -> float:
        return self._get_idle_current(self.sensor_config.inter_frame_idle_state)

    @property
    def sweep_active_charge(self) -> float:
        return self.sweep_active_duration * self.active_current

    @property
    def sweep_idle_charge(self) -> float:
        return self.sweep_idle_duration * self.inter_sweep_idle_current

    @property
    def sweep_charge(self) -> float:
        return self.sweep_active_charge + self.sweep_idle_charge

    @property
    def frame_active_charge(self) -> float:
        return self.sweep_charge * self.spf

    @property
    def frame_idle_charge(self) -> float:
        return self.frame_idle_duration * self.inter_frame_idle_current

    @property
    def frame_charge(self) -> float:
        return self.frame_active_charge + self.frame_idle_charge

    @property
    def average_current(self) -> float:
        return self.frame_charge / self.frame_period

    @staticmethod
    def _get_idle_current(idle_state: IdleState) -> float:
        return {
            IdleState.READY: 70.0e-3,
            IdleState.SLEEP: 1.8e-3,
            IdleState.DEEP_SLEEP: 1.2e-3,
        }[idle_state]
